/*
 * Copyright 2010 Sergey Martynov http://studiomobile.ru
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.studiomobile.http;

import org.apache.http.*;
import org.apache.http.impl.DefaultConnectionReuseStrategy;
import org.apache.http.impl.nio.DefaultClientIOEventDispatch;
import org.apache.http.impl.nio.reactor.DefaultConnectingIOReactor;
import org.apache.http.nio.NHttpClientConnection;
import org.apache.http.nio.protocol.BufferingHttpClientHandler;
import org.apache.http.nio.reactor.*;
import org.apache.http.params.BasicHttpParams;
import org.apache.http.params.DefaultedHttpParams;
import org.apache.http.params.HttpParams;
import org.apache.http.protocol.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InterruptedIOException;
import java.net.InetSocketAddress;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static ru.studiomobile.http.HttpClientConfig.DISPATCHING_THREADS;

public class BasicHttpClient extends HttpClientHelper implements HttpClient {

    private final static Logger log = LoggerFactory.getLogger(BasicHttpClient.class);

    private final HttpParams params;
    private final ConnectingIOReactor ioReactor;
    private final SesionRequestHandler sesionRequestHandler;
    private final Map<HttpHost, InetSocketAddress> resolvedHosts;

    public BasicHttpClient() throws IOException {
        this(new BasicHttpParams());
    }

    public BasicHttpClient(final HttpParams config) throws IOException {
        params = new DefaultedHttpParams(config, HttpClientConfig.defaults());
        ioReactor = new DefaultConnectingIOReactor(params.getIntParameter(DISPATCHING_THREADS, 1), params);
        sesionRequestHandler = new SesionRequestHandler();
        resolvedHosts = new HashMap<HttpHost, InetSocketAddress>();

        BasicHttpProcessor processor = new BasicHttpProcessor();
        processor.addInterceptor(new RequestContent());
        processor.addInterceptor(new RequestTargetHost());
        processor.addInterceptor(new RequestConnControl());
        processor.addInterceptor(new RequestUserAgent());
        processor.addInterceptor(new RequestExpectContinue());

        HttpRequestExecutionHandler executionHandler = new HttpRequestExecutionHandler(new RedirectHandler());

        BufferingHttpClientHandler protocolHandler = new BufferingHttpClientHandler(
                processor,
                executionHandler,
                new DefaultConnectionReuseStrategy(),
                params);

        HttpClientConnectionHandler clientConnectionHandler = new HttpClientConnectionHandler(protocolHandler);

        final IOEventDispatch ioEventDispatch = new DefaultClientIOEventDispatch(clientConnectionHandler, params);

        Thread webThread = new Thread(new Runnable() {
            public void run() {
                try {
                    ioReactor.execute(ioEventDispatch);
                } catch (InterruptedIOException ex) {
                    log.error("Web thread interrupted.", ex);
                } catch (IOException e) {
                    log.error("I/O error in web thread.", e);
                }
                log.debug("Shutdown web thread.");
            }
        });
        webThread.start();
    }


    public IOReactorStatus getStatus() {
        return ioReactor.getStatus();
    }

    public void shutdown(long waitMs) throws IOException {
        ioReactor.shutdown(waitMs);
    }

    public void shutdown() throws IOException {
        ioReactor.shutdown();
    }

    protected void finalize() throws Throwable {
        shutdown();
        super.finalize();
    }


    public HttpRequestFuture executeRequest(InetSocketAddress address, HttpRequest request) {
        HttpRequestFutureImpl future = new HttpRequestFutureImpl(request, params);
        ioReactor.connect(address, null, future, sesionRequestHandler);
        return future;
    }

    public HttpRequestFuture head(URL url, Header... headers) {
        return executeRequest(getAddress(url), buildRequest("HEAD", url, headers));
    }

    public HttpRequestFuture get(URL url, Header... headers) {
        return executeRequest(getAddress(url), buildRequest("GET", url, headers));
    }

    public HttpRequestFuture post(URL url, HttpEntity entity, Header... headers) {
        return executeRequest(getAddress(url), buildRequest("POST", url, entity, headers));
    }

    public HttpRequestFuture put(URL url, HttpEntity entity, Header... headers) {
        return executeRequest(getAddress(url), buildRequest("PUT", url, entity, headers));
    }

    public HttpRequestFuture delete(URL url, Header... headers) {
        return executeRequest(getAddress(url), buildRequest("DELETE", url, headers));
    }


    private HttpHost getHost(HttpRequest request) throws MalformedURLException {
        URL url = new URL(request.getRequestLine().getUri());
        return new HttpHost(url.getHost(), getPort(url), url.getProtocol());
    }


    private class RedirectHandler implements HttpRequestExecutionHandler.RedirectHandler {

        public void handleRedirect(HttpRequestFutureImpl state, HttpContext context) throws IOException {
            NHttpClientConnection connection = (NHttpClientConnection) context.getAttribute(ExecutionContext.HTTP_CONNECTION);
            List<HttpRequest> redirects = state.getRedirects();
            HttpRequest initial = redirects.size() > 1 ? redirects.get(redirects.size() - 2) : state.getRequest();
            HttpRequest redirect = redirects.get(redirects.size() - 1);
            HttpHost initialHost = getHost(initial);
            HttpHost redirectHost = getHost(redirect);
            log.debug("Redirected to {} from {}", redirectHost, initial);
            if (initialHost.equals(redirectHost) || haveSameAddress(initialHost, redirectHost)) {
                try {
                    connection.submitRequest(redirect);
                } catch (HttpException e) {
                    state.failed(e);
                }
            } else {
                state.setActiveConnection(null);
                connection.close();
                ioReactor.connect(new InetSocketAddress(redirectHost.getHostName(), redirectHost.getPort()), null, state, sesionRequestHandler);
            }
        }

        private boolean haveSameAddress(HttpHost h1, HttpHost h2) {
            return resolvedHosts.containsKey(h1) && resolvedHosts.get(h1).equals(resolvedHosts.get(h2));
        }
    }

    private class SesionRequestHandler implements SessionRequestCallback {

        public void completed(SessionRequest request) {
            HttpRequestFutureImpl state = getState(request);
            try {
                resolvedHosts.put(getHost(state.getActiveRequest()), (InetSocketAddress) request.getRemoteAddress());
            } catch (MalformedURLException ignore) {
            }
        }

        public void failed(SessionRequest request) {
            getState(request).failed(request.getException());
        }

        public void timeout(SessionRequest request) {
            getState(request).failed(new ConnectionClosedException("Connection request timed out"));
        }

        public void cancelled(SessionRequest request) {
            getState(request).cancel(false);
        }

        private HttpRequestFutureImpl getState(SessionRequest request) {
            return (HttpRequestFutureImpl) request.getAttachment();
        }
    }
}
