/*
 * Copyright 2010 Sergey Martynov http://studiomobile.ru
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.studiomobile.http;

import org.apache.http.HttpRequest;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.message.BasicHttpRequest;
import org.apache.http.nio.NHttpClientConnection;
import org.apache.http.params.HttpParams;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

class HttpRequestFutureImpl implements HttpRequestFuture {
    private final HttpRequest request;
    private final HttpParams params;
    private final List<HttpRequest> redirects;
    private volatile HttpResponse response;
    private volatile Exception error;
    private volatile HttpClientListener listener;
    private volatile boolean cancelled = false;
    private volatile boolean failed = false;
    private NHttpClientConnection activeConnection;

    public HttpRequestFutureImpl(HttpRequest request, HttpParams params) {
        this.request = request;
        this.params = params;
        this.redirects = new ArrayList<HttpRequest>();
    }

    public HttpRequest getRequest() {
        return request;
    }

    public HttpRequest getActiveRequest() {
        return redirects.size() > 0 ? redirects.get(redirects.size() - 1) : request;
    }

    public boolean isDone() {
        return response != null || cancelled || failed;
    }

    public void finished(HttpResponse response) {
        synchronized (this) {
            this.response = response;
            notifyAll();
        }
        if (listener != null) {
            listener.requestFinished(this, response);
        }
    }

    public HttpResponse get() throws InterruptedException, ExecutionException {
        synchronized (this) {
            if (!isDone()) {
                wait();
            }
        }
        return response;
    }

    public HttpResponse get(long l, TimeUnit timeUnit) throws InterruptedException, ExecutionException, TimeoutException {
        synchronized (this) {
            if (!isDone()) {
                wait(timeUnit.toMillis(l), (int) timeUnit.toNanos(l));
            }
        }
        return response;
    }

    public List<HttpRequest> getRedirects() {
        return redirects;
    }

    public boolean redirect(HttpResponse response) {
        if (isDone()) return false;
        if (!params.getBooleanParameter(HttpClientConfig.HANDLE_REDIRECTS, false)) return false;
        int maxRedirects = params.getIntParameter(HttpClientConfig.MAX_REDIRECTS, -1);
        if (redirects.size() >= maxRedirects && maxRedirects >= 0) return false;

        HttpRequest activeRequest = getActiveRequest();
        int statusCode = response.getStatusLine().getStatusCode();
        String method = statusCode == HttpStatus.SC_SEE_OTHER ? "GET" : activeRequest.getRequestLine().getMethod();
        String location = response.getFirstHeader(HttpHeaders.LOCATION).getValue();
        BasicHttpRequest redirect = new BasicHttpRequest(method, location);
        redirect.setHeaders(activeRequest.getAllHeaders());
        redirects.add(redirect);
        return true;
    }

    public boolean isCancelled() {
        return cancelled;
    }

    public boolean cancel(boolean mayInterrupt) {
        synchronized (this) {
            cancelled = true;
            notifyAll();
        }
        if (listener != null) {
            listener.requestCancelled(this);
        }
        return true;
    }

    public Exception getError() {
        return error;
    }

    public void failed(Exception exception) {
        synchronized (this) {
            error = exception;
            failed = true;
            notifyAll();
        }
        if (listener != null) {
            listener.requestFailed(this);
        }
    }

    public HttpClientListener getListener() {
        return listener;
    }

    public void setListener(HttpClientListener listener) {
        this.listener = listener;
        if (listener != null) {
            if (response != null) {
                listener.requestFinished(this, response);
            }
            if (cancelled) {
                listener.requestCancelled(this);
                return;
            }
            if (failed) {
                listener.requestFailed(this);
            }
        }
    }

    public void setActiveConnection(NHttpClientConnection connection) {
        activeConnection = connection;
    }

    public NHttpClientConnection getActiveConnection() {
        return activeConnection;
    }
}
