/*
 * Copyright 2010 Sergey Martynov http://studiomobile.ru
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.studiomobile.http.entity.nio;

import org.apache.http.entity.mime.content.ContentBody;
import org.apache.http.entity.mime.content.FileBody;
import org.apache.http.nio.ContentEncoder;
import org.apache.http.nio.ContentEncoderChannel;
import org.apache.http.nio.FileContentEncoder;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;

public class NContentBodyWriter {

    private final ByteBuffer buffer;

    public NContentBodyWriter(ByteBuffer buffer) {
        this.buffer = buffer;
    }

    public static NContentBodyWriter createWriter(final ContentBody body) throws IOException {
        if (body instanceof FileBody) {
            return new FileWriter(((FileBody)body).getFile());
        }
        return new NContentBodyWriter(encodeBody(body));
    }

    public boolean encodeTo(ContentEncoder encoder) throws IOException {
        encoder.write(buffer);
        return !buffer.hasRemaining();
    }

    public static ByteBuffer encodeBody(ContentBody body) throws IOException {
        long contentLength = body.getContentLength();
        if (contentLength <= 0) {
            throw new IllegalArgumentException("NContentBodyWriter does not support bodies without known content length.");
        }
        if (contentLength > Integer.MAX_VALUE) {
            throw new IllegalArgumentException("NContentBodyWriter does not support bodies with very large content length.");
        }
        ByteBuffer buffer = ByteBuffer.allocate((int) contentLength);
        body.writeTo(new ByteBufferOutputStream(buffer));
        buffer.flip();
        return buffer;
    }

    public void close() {}

    static class ByteBufferOutputStream extends OutputStream {
        private final ByteBuffer buffer;

        public ByteBufferOutputStream(ByteBuffer buffer) {
            this.buffer = buffer;
        }

        public void write(int i) throws IOException {
            buffer.put((byte) i);
        }

        public void write(byte[] bytes) throws IOException {
            buffer.put(bytes);
        }

        public void write(byte[] bytes, int offset, int length) throws IOException {
            buffer.put(bytes, offset, length);
        }
    }

    private static class FileWriter extends NContentBodyWriter {
        private final File file;
        private FileChannel fileChannel;
        private long written;

        public FileWriter(File file) {
            super(null);
            this.file = file;
        }

        public boolean encodeTo(ContentEncoder encoder) throws IOException {
            if (fileChannel == null) {
                FileInputStream in = new FileInputStream(file);
                fileChannel = in.getChannel();
                written = 0;
            }

            long transferred;
            if (encoder instanceof FileContentEncoder) {
                transferred = ((FileContentEncoder)encoder).transfer(fileChannel, written, Long.MAX_VALUE);
            } else {
                transferred = fileChannel.transferTo(written, Long.MAX_VALUE, new ContentEncoderChannel(encoder));
            }

            if (transferred > 0) {
                written += transferred;
            }

            return written >= fileChannel.size();
        }

        public void close() {
            if (fileChannel != null) {
                try {
                    fileChannel.close();
                } catch (IOException ignored) {
                }
                fileChannel = null;
            }
        }
    }
}
