/*
 * Copyright 2010 Sergey Martynov http://studiomobile.ru
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.studiomobile.http.entity.nio;

import org.apache.http.entity.mime.*;
import org.apache.http.entity.mime.content.ContentBody;
import org.apache.http.message.BasicHeader;
import org.apache.http.nio.ContentEncoder;
import org.apache.http.nio.IOControl;
import org.apache.http.nio.entity.ProducingNHttpEntity;
import org.apache.http.protocol.HTTP;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.util.LinkedList;
import java.util.List;

public class NMultipartEntity extends NMultipartEntityUtils implements ProducingNHttpEntity {

    private final HttpMultipart multipart;
    private final org.apache.http.Header contentType;

    private long length;
    private volatile boolean dirty; // used to decide whether to recalculate length

    private final ByteBuffer boundary;
    private List<ByteBuffer> pendingBuffers;
    private List<FormBodyPart> pendingParts;
    private NContentBodyWriter partBodyWriter;

    public NMultipartEntity(HttpMultipartMode mode, String boundary, Charset charset) {
        this.multipart = new HttpMultipart("form-data", charset != null ? charset : MIME.DEFAULT_CHARSET, boundary != null ? boundary : generateBoundary());
        this.multipart.setMode(mode != null ? mode : HttpMultipartMode.STRICT);
        this.contentType = new BasicHeader(HTTP.CONTENT_TYPE, generateContentType(multipart.getBoundary(), multipart.getCharset()));
        this.dirty = true;
        this.boundary = encode(multipart.getBoundary(), multipart.getCharset());
    }

    public NMultipartEntity(HttpMultipartMode mode) {
        this(mode, null, null);
    }

    public NMultipartEntity() {
        this(HttpMultipartMode.STRICT, null, null);
    }

    public void addPart(final FormBodyPart bodyPart) {
        this.multipart.addBodyPart(bodyPart);
        this.dirty = true;
    }

    public void addPart(final String name, final ContentBody contentBody) {
        addPart(new FormBodyPart(name, contentBody));
    }

    public void produceContent(ContentEncoder encoder, IOControl ioctrl) throws IOException {
        if (pendingBuffers == null) {
            pendingBuffers = new LinkedList<ByteBuffer>();
            pendingParts = new LinkedList<FormBodyPart>(multipart.getBodyParts());
        }

        while (pendingBuffers.size() > 0 && !encoder.isCompleted()) {
            ByteBuffer buffer = pendingBuffers.get(0);
            encoder.write(buffer);
            if (buffer.hasRemaining()) return;
            pendingBuffers.remove(0);
            buffer.rewind();
        }

        if (encoder.isCompleted()) return;

        if (partBodyWriter != null) {
            boolean done = partBodyWriter.encodeTo(encoder);
            if (done) {
                pendingBuffers.add(CR_LF);
                partBodyWriter.close();
                partBodyWriter = null;
            }
        }

        if (partBodyWriter == null) {
            if (pendingParts.size() > 0) {
                pendingBuffers.add(TWO_DASHES);
                pendingBuffers.add(boundary);
                pendingBuffers.add(CR_LF);

                FormBodyPart part = pendingParts.remove(0);
                encodePartHeader(part, multipart.getMode(), multipart.getCharset(), pendingBuffers);
                partBodyWriter = NContentBodyWriter.createWriter(part.getBody());

            } else {
                pendingBuffers.add(TWO_DASHES);
                pendingBuffers.add(boundary);
                pendingBuffers.add(TWO_DASHES);
                pendingBuffers.add(CR_LF);
            }
        }

        if (pendingBuffers.size() > 0 || pendingParts.size() > 0 || partBodyWriter != null) {
            produceContent(encoder, ioctrl);
        } else {
            finish();
            encoder.complete();
        }
    }

    public void finish() throws IOException {
        pendingBuffers = null;
        pendingParts = null;
        if (partBodyWriter != null) {
            partBodyWriter.close();
        }
        partBodyWriter = null;
    }

    public void writeTo(OutputStream outstream) throws IOException {
        this.multipart.writeTo(outstream);
    }

    public boolean isChunked() {
        return !isRepeatable();
    }

    public boolean isStreaming() {
        return !isRepeatable();
    }

    public boolean isRepeatable() {
        for (FormBodyPart part: this.multipart.getBodyParts()) {
            ContentBody body = part.getBody();
            if (body.getContentLength() < 0) {
                return false;
            }
        }
        return true;
    }

    public long getContentLength() {
        if (this.dirty) {
            this.length = this.multipart.getTotalLength();
            this.dirty = false;
        }
        return this.length;
    }

    public org.apache.http.Header getContentType() {
        return this.contentType;
    }

    public org.apache.http.Header getContentEncoding() {
        return null;
    }

    public InputStream getContent() throws IOException, IllegalStateException {
        throw new UnsupportedOperationException("NMultipartEntity does not implement #getContent()");
    }

    public void consumeContent() throws IOException {
    }
}
